"""Basic CP decomposition script for processing CP-CPMG data sets
   Data is input as a JCAMP dx file exported from Bruker Topspin software   
   
   Most packages are included in the scipy python package.
   The scikit tensor package must be downloaded seperately  
   
   Harris Mason
   LLNL 
   4/10/18 """

import numpy as np, pylab as plt
import re
from sktensor import dtensor, cp_als

# sine-bell apodization
def apod(end, off, size, power):
    y = np.power(np.sin(np.pi * off + np.pi * (end - off) * np.arange(size) / (size -1)), power)
    return y

text = open('DGA CPMG 4mm rev 400v2.dx', 'r')

### Code to read in the .dx file generated by Bruker topspin and spit out the individual FIDs
count = 0 #for record keeping below
iters = [0,0] #for knowing how many FIDs and how many points
while 1:
    line = text.readline()
    dim_line = re.match('##VAR_DIM=', line) #finds the line with the total number of FIDs and number of points for each FID
    if dim_line:
        iters = [int(line.split()[1].replace(',','')),
                 int(line.split()[2].replace(',',''))]
        dt_table = np.zeros((iters[0],iters[1]*2), dtype = np.int) #generates the array for data storage
    dt_start = re.match('##DATA TABLE=', line) #finds the start of each subsequent data set
    ### Next few lines find the data and store real/imaginary as interleaved data
    if dt_start:
        if count % 2 == 0:
            for i in range(iters[1]):
                line = text.readline().split()
                dt_table[count/2,i*2] = int(line[1])
        if count % 2 != 0:
            for j in range(iters[1]):
                line = text.readline().split()
                dt_table[count/2,j*2+1] = int(line[1])
        count +=1
    if line == '':
        break
text.close()

#user input of the number of number of real/imaginary points per echo and number of echoes
its = 602 #number of real and imaginary points for each echo
nech = 99 #number of echos collected

tau = 6  #cpmg echo delay in ms
T2_decay_time = tau*np.arange(nech)

ct = np.array([0.5,0.7,1,1.5,2,3,4,5,7,10,12,15]) #ct array used

storage = np.zeros([iters[0],its, nech]) #storage array for the final 3D matrix

plt.figure()
for ii in range(iters[0]):
    temp = dt_table[ii][300+2:]  #The program only uses the echoes. The FID before the first echo is discarded.  
    for k in range(99):
        t2 = temp[k*its:(k+1)*its]
        storage[ii,:,k] = t2

T = dtensor(np.rot90(storage))  #this transforms the 3D matrix into a tensor where all the work is done
# The matrix is rotated such that the z-axis is the shortest axis
ncomp = 2 #user defined number of components to use. May need to increase or decrease depending on fitting results
P, fit, itr, exectimes = cp_als(T, ncomp, init = 'nvecs') #the cp-als function from the ski-tensor 'nvecs' denotes using HSVOD to initialize

#plot all of the raw results of the decomposition in a multi-panel figure


plt.subplot(221)
plt.title('Extracted FIDs') #this plot shows the interleaved FIDS
plt.xlabel('pts', fontsize = 16)
plt.plot(P.U[0])  #(P.U[0][::2]) will plot real portion and (P.U[0][1::2] will plot imaginary portion of the FIDS 
plt.subplot(222)
plt.title('Intensity vs Contact Time')
plt.xlabel('time (ms)', fontsize = 16)
plt.plot(ct, P.U[1]) #contact time curves
plt.subplot(223)
plt.title('T2 decay curves')
plt.xlabel('time(ms)', fontsize = 16)
plt.plot(T2_decay_time, P.U[2]) #T2 curves

# code below performs a FT on the data and displays magnitude data of each component
plt.subplot(224)
plt.title('Spectral Components')
plt.xlabel('Chemical shift (ppm)', fontsize = 16)
final = np.zeros((2048, ncomp)).astype('complex')
cdat = P.U[0][::2] + 1j*P.U[0][1::2] #places the data into a complex data set fpr the FT
inten = np.array([0.69+0.38, 1.15]) #the intensity of the components are back corrected based on external fits of the T2 data
delt = 0.5; start = 435.9; end = start-(delt*2048) #values to generate chemical shift scale.  Values are extracted from the Bruker file
x_axis = np.linspace(end, start, 2048) #final x-axis
for jj in range(ncomp):
    cdn = cdat[:,jj][:-1]
    size = cdn.shape[-1]
    y = apod(.5, .1, size, 1).astype(cdn.dtype) #apodizization array for the data
    cdat2 = np.zeros((2048)).astype(cdn.dtype)
    cdat2[1024-(301/2):1024+(301/2)] = y*cdn  #places the echo in the center of new array effectively zero filling on both sides
    final[:,jj] = inten[jj]*np.fft.fftshift(np.fft.fft(cdat2))
plt.plot(x_axis, np.sqrt((final.real)**2+(final.imag)**2)) #plot the magnitude data. 
plt.xlim(np.max(x_axis), np.min(x_axis)) #reverses the x axis

rec = np.dot(P.U[1], final.T).T #full 2D data set reconstructed

num = 8 #user chosen data set to export
mag =  np.sqrt(rec[:,num].real**2+rec[:,num].imag**2) #magnitude transform
out_pfc = np.vstack((x_axis, mag)).T #combines into two column array for export
np.savetxt('filename.txt', out_pfc) #export the file as two column .txt file for plotting

plt.figure()
plt.title('final reconstructed data for spectrum %d' %num)
plt.xlabel('Chemical Shift (ppm)', fontsize = 16)
plt.plot(x_axis, mag) #plot final data
plt.xlim(np.max(x_axis), np.min(x_axis)) #reverses the x axis
plt.show()